@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,6 +21,7 @@
 
 # This contains various Displayables that handle events.
 
+from __future__ import print_function
 
 import renpy.display
 import renpy.audio
@@ -31,6 +32,7 @@ import pygame_sdl2 as pygame
 
 import math
 
+
 def compile_event(key, keydown):
     """
     Compiles a keymap entry into a python expression.
@@ -87,31 +89,41 @@ def compile_event(key, keydown):
     while part[0] in MODIFIERS:
         modifiers.add(part.pop(0))
 
+    key = "_".join(part)
+
     if "repeat" in modifiers:
         rv += " and (ev.repeat)"
     else:
         rv += " and (not ev.repeat)"
 
-    if "alt" in modifiers:
-        rv += " and (ev.mod & %d)" % pygame.KMOD_ALT
-    else:
-        rv += " and not (ev.mod & %d)" % pygame.KMOD_ALT
+    if key not in [ "K_LALT", "K_RALT" ]:
 
-    if "meta" in modifiers:
-        rv += " and (ev.mod & %d)" % pygame.KMOD_META
-    else:
-        rv += " and not (ev.mod & %d)" % pygame.KMOD_META
+        if "alt" in modifiers:
+            rv += " and (ev.mod & %d)" % pygame.KMOD_ALT
+        else:
+            rv += " and not (ev.mod & %d)" % pygame.KMOD_ALT
 
-    if "ctrl" in modifiers:
-        rv += " and (ev.mod & %d)" % pygame.KMOD_CTRL
-    else:
-        rv += " and not (ev.mod & %d)" % pygame.KMOD_CTRL
+    if key not in [ "K_LGUI", "K_RGUI" ]:
 
-    if "shift" in modifiers:
-        rv += " and (ev.mod & %d)" % pygame.KMOD_SHIFT
+        if "meta" in modifiers:
+            rv += " and (ev.mod & %d)" % pygame.KMOD_META
+        else:
+            rv += " and not (ev.mod & %d)" % pygame.KMOD_META
 
-    if "noshift" in modifiers:
-        rv += " and not (ev.mod & %d)" % pygame.KMOD_SHIFT
+    if key not in [ "K_LCTRL", "K_RCTRL" ]:
+
+        if "ctrl" in modifiers:
+            rv += " and (ev.mod & %d)" % pygame.KMOD_CTRL
+        else:
+            rv += " and not (ev.mod & %d)" % pygame.KMOD_CTRL
+
+    if key not in [ "K_LSHIFT", "K_RSHIFT" ]:
+
+        if "shift" in modifiers:
+            rv += " and (ev.mod & %d)" % pygame.KMOD_SHIFT
+
+        if "noshift" in modifiers:
+            rv += " and not (ev.mod & %d)" % pygame.KMOD_SHIFT
 
     if len(part) == 1:
         if len(part[0]) != 1:
@@ -129,8 +141,6 @@ def compile_event(key, keydown):
             else:
                 return "(False)"
 
-        key = "_".join(part)
-
         rv += " and ev.key == %d)" % (getattr(pygame.constants, key))
 
     return rv
@@ -185,6 +195,7 @@ def queue_event(name, up=False, **kwargs):
     ev = pygame.event.Event(renpy.display.core.EVENTNAME, data)
     pygame.event.post(ev)
 
+
 def map_event(ev, keysym):
     """
     :doc: udd_utility
@@ -212,6 +223,7 @@ def map_event(ev, keysym):
 
     return check_code(ev)
 
+
 def map_keyup(ev, name):
     """Returns true if the event matches the named keycode being released."""
 
@@ -243,7 +255,7 @@ def skipping(ev):
         renpy.config.skipping = "slow"
         renpy.exports.restart_interaction()
 
-    if map_keyup(ev, "skip"):
+    if map_keyup(ev, "skip") or map_event(ev, "stop_skipping"):
         renpy.config.skipping = None
         renpy.exports.restart_interaction()
 
@@ -272,19 +284,27 @@ def predict_action(var):
         for i in var:
             predict_action(i)
 
-def run(var, *args, **kwargs):
+
+def run(action, *args, **kwargs):
     """
-    Runs a variable. This is done by calling all the functions, and
-    iterating over the lists and tuples.
+    :doc: run
+    :name: renpy.run
+    :args: (action)
+
+    Run an action or list of actions. A single action is called with no
+    arguments, a list of actions is run in order using this function, and
+    None is ignored.
+
+    Returns the result of the first action to return a value.
     """
 
-    if var is None:
+    if action is None:
         return None
 
-    if isinstance(var, (list, tuple)):
+    if isinstance(action, (list, tuple)):
         rv = None
 
-        for i in var:
+        for i in action:
             new_rv = run(i, *args, **kwargs)
 
             if new_rv is not None:
@@ -292,7 +312,8 @@ def run(var, *args, **kwargs):
 
         return rv
 
-    return var(*args, **kwargs)
+    return action(*args, **kwargs)
+
 
 def run_unhovered(var):
     """
@@ -312,6 +333,7 @@ def run_unhovered(var):
     if f is not None:
         f()
 
+
 def run_periodic(var, st):
 
     if isinstance(var, (list, tuple)):
@@ -328,33 +350,45 @@ def run_periodic(var, st):
     if isinstance(var, renpy.ui.Action):
         return var.periodic(st)
 
-def is_selected(clicked):
 
-    if isinstance(clicked, (list, tuple)):
-        for i in clicked:
-            if isinstance(i, renpy.store.SelectedIf): # @UndefinedVariable
+def is_selected(action):
+    """
+    :doc: run
+
+    Returns true if `action` indicates it is selected, or false otherwise.
+    """
+
+    if isinstance(action, (list, tuple)):
+        for i in action:
+            if isinstance(i, renpy.store.SelectedIf):  # @UndefinedVariable
                 return i.get_selected()
-        return any(is_selected(i) for i in clicked)
+        return any(is_selected(i) for i in action)
 
-    elif isinstance(clicked, renpy.ui.Action):
-        return clicked.get_selected()
+    elif isinstance(action, renpy.ui.Action):
+        return action.get_selected()
     else:
         return False
 
 
-def is_sensitive(clicked):
+def is_sensitive(action):
+    """
+    :doc: run
 
-    if isinstance(clicked, (list, tuple)):
-        for i in clicked:
-            if isinstance(i, renpy.store.SensitiveIf): # @UndefinedVariable
+    Returns true if `action` indicates it is sensitive, or False otherwise.
+    """
+
+    if isinstance(action, (list, tuple)):
+        for i in action:
+            if isinstance(i, renpy.store.SensitiveIf):  # @UndefinedVariable
                 return i.get_sensitive()
-        return all(is_sensitive(i) for i in clicked)
+        return all(is_sensitive(i) for i in action)
 
-    elif isinstance(clicked, renpy.ui.Action):
-        return clicked.get_sensitive()
+    elif isinstance(action, renpy.ui.Action):
+        return action.get_sensitive()
     else:
         return True
 
+
 def alt(clicked):
 
     if isinstance(clicked, (list, tuple)):
@@ -378,6 +412,7 @@ def alt(clicked):
 ##############################################################################
 # Special-Purpose Displayables
 
+
 class Keymap(renpy.display.layout.Null):
     """
     This is a behavior that maps keys to actions that are called when
@@ -396,10 +431,10 @@ class Keymap(renpy.display.layout.Null):
     def event(self, ev, x, y, st):
 
         for name, action in self.keymap.iteritems():
+
             if map_event(ev, name):
 
-                if self.style.activate_sound:
-                    renpy.audio.music.play(self.style.activate_sound, channel="sound")
+                renpy.exports.play(self.style.activate_sound)
 
                 rv = run(action)
 
@@ -433,16 +468,24 @@ class PauseBehavior(renpy.display.layout.Null):
     return a value after a certain amount of time has elapsed.
     """
 
-    def __init__(self, delay, result=False, **properties):
+    voice = False
+
+    def __init__(self, delay, result=False, voice=False, **properties):
         super(PauseBehavior, self).__init__(**properties)
 
         self.delay = delay
         self.result = result
+        self.voice = voice
 
     def event(self, ev, x, y, st):
 
         if st >= self.delay:
 
+            if self.voice and renpy.config.nw_voice:
+                if (not renpy.config.afm_callback()) or renpy.display.tts.is_active():
+                    renpy.game.interface.timeout(0.05)
+                    return
+
             # If we have been drawn since the timeout, simply return
             # true. Otherwise, force a redraw, and return true when
             # it comes back.
@@ -451,9 +494,9 @@ class PauseBehavior(renpy.display.layout.Null):
             else:
                 renpy.game.interface.force_redraw = True
 
-
         renpy.game.interface.timeout(max(self.delay - st, 0))
 
+
 class SoundStopBehavior(renpy.display.layout.Null):
     """
     This is a class implementing the sound stop behavior,
@@ -467,7 +510,6 @@ class SoundStopBehavior(renpy.display.layout.Null):
         self.channel = channel
         self.result = result
 
-
     def event(self, ev, x, y, st):
 
         if not renpy.audio.music.get_playing(self.channel):
@@ -485,8 +527,11 @@ class SayBehavior(renpy.display.layout.Null):
     """
 
     focusable = True
+    text = None
 
-    def __init__(self, default=True, afm=None, dismiss=[ 'dismiss' ], allow_dismiss=None, **properties):
+    dismiss_unfocused = [ 'dismiss_unfocused' ]
+
+    def __init__(self, default=True, afm=None, dismiss=[ 'dismiss' ], allow_dismiss=None, dismiss_unfocused=[ 'dismiss_unfocused' ], **properties):
         super(SayBehavior, self).__init__(default=default, **properties)
 
         if not isinstance(dismiss, (list, tuple)):
@@ -505,8 +550,9 @@ class SayBehavior(renpy.display.layout.Null):
     def _tts_all(self):
         raise renpy.display.tts.TTSRoot()
 
-    def set_afm_length(self, afm_length):
-        self.afm_length = max(afm_length, 1)
+    def set_text(self, text):
+        self.text = text
+        self.afm_length = max(text.end - text.start, 1)
 
     def event(self, ev, x, y, st):
 
@@ -514,12 +560,12 @@ class SayBehavior(renpy.display.layout.Null):
 
             afm_delay = ( 1.0 * ( renpy.config.afm_bonus + self.afm_length ) / renpy.config.afm_characters ) * renpy.game.preferences.afm_time
 
-            if renpy.game.preferences.text_cps:
-                afm_delay += 1.0 / renpy.game.preferences.text_cps * self.afm_length
+            if self.text is not None:
+                afm_delay += self.text.get_time()
 
             if st > afm_delay:
                 if renpy.config.afm_callback:
-                    if renpy.config.afm_callback():
+                    if renpy.config.afm_callback() and not renpy.display.tts.is_active():
                         return True
                     else:
                         renpy.game.interface.timeout(0.1)
@@ -528,18 +574,46 @@ class SayBehavior(renpy.display.layout.Null):
             else:
                 renpy.game.interface.timeout(afm_delay - st)
 
-        for dismiss in self.dismiss:
+        dismiss = [ (i, True) for i in self.dismiss ] + [ (i, False) for i in self.dismiss_unfocused ]
 
-            if map_event(ev, dismiss) and self.is_focused():
+        for dismiss_event, check_focus in dismiss:
+
+            if map_event(ev, dismiss_event):
+
+                if check_focus and not self.is_focused():
+                    continue
 
                 if renpy.config.skipping:
                     renpy.config.skipping = None
                     renpy.exports.restart_interaction()
                     raise renpy.display.core.IgnoreEvent()
 
+                if not renpy.config.enable_rollback_side:
+                    rollback_side = "disable"
+                if renpy.mobile:
+                    rollback_side = renpy.game.preferences.mobile_rollback_side
+                else:
+                    rollback_side = renpy.game.preferences.desktop_rollback_side
+
+                if ev.type == pygame.MOUSEBUTTONUP:
+
+                    percent = 1.0 * x / renpy.config.screen_width
+
+                    if rollback_side == "left":
+
+                        if percent < renpy.config.rollback_side_size:
+                            renpy.exports.rollback()
+                            raise renpy.display.core.IgnoreEvent()
+
+                    elif rollback_side == "right":
+
+                        if (1.0 - percent) < renpy.config.rollback_side_size:
+                            renpy.exports.rollback()
+                            raise renpy.display.core.IgnoreEvent()
+
                 if renpy.game.preferences.using_afm_enable and \
-                    renpy.game.preferences.afm_enable and \
-                    not renpy.game.preferences.afm_after_click:
+                        renpy.game.preferences.afm_enable and \
+                        not renpy.game.preferences.afm_after_click:
 
                     renpy.game.preferences.afm_enable = False
                     renpy.exports.restart_interaction()
@@ -562,16 +636,27 @@ class SayBehavior(renpy.display.layout.Null):
                     return True
                 elif renpy.game.context().seen_current(True):
                     return True
+                else:
+                    renpy.config.skipping = False
+                    renpy.exports.restart_interaction()
+
             else:
                 renpy.game.interface.timeout(skip_delay - st)
 
-
         return None
 
 
 ##############################################################################
 # Button
 
+KEY_EVENTS = (
+    pygame.KEYDOWN,
+    pygame.KEYUP,
+    pygame.TEXTEDITING,
+    pygame.TEXTINPUT
+    )
+
+
 class Button(renpy.display.layout.Window):
 
     keymap = { }
@@ -584,9 +669,13 @@ class Button(renpy.display.layout.Window):
 
     role_parameter = None
 
+    keysym = None
+    alternate_keysym = None
+
     def __init__(self, child=None, style='button', clicked=None,
                  hovered=None, unhovered=None, action=None, role=None,
                  time_policy=None, keymap={}, alternate=None,
+                 selected=None, sensitive=None, keysym=None, alternate_keysym=None,
                  **properties):
 
         if isinstance(clicked, renpy.ui.Action):
@@ -595,17 +684,28 @@ class Button(renpy.display.layout.Window):
         super(Button, self).__init__(child, style=style, **properties)
 
         self.action = action
+        self.selected = selected
+        self.sensitive = sensitive
         self.clicked = clicked
         self.hovered = hovered
         self.unhovered = unhovered
         self.alternate = alternate
 
-        self.focusable = True #(clicked is not None) or (action is not None)
+        self.focusable = True  # (clicked is not None) or (action is not None)
         self.role_parameter = role
+
         self.keymap = keymap
 
+        self.keysym = keysym
+        self.alternate_keysym = alternate_keysym
+
         self.time_policy_data = None
 
+        self._duplicatable = False
+
+    def _duplicate(self, args):
+        return self
+
     def predict_one_action(self):
         predict_action(self.clicked)
         predict_action(self.hovered)
@@ -689,16 +789,25 @@ class Button(renpy.display.layout.Window):
         if self.child is not None:
             self.child.set_transform_event(self.role + "idle")
 
+    def is_selected(self):
+        if self.selected is not None:
+            return self.selected
+        return is_selected(self.action)
+
+    def is_sensitive(self):
+        if self.sensitive is not None:
+            return self.sensitive
+        return is_sensitive(self.action)
 
     def per_interact(self):
 
         if self.action is not None:
-            if is_selected(self.action):
+            if self.is_selected():
                 role = 'selected_'
             else:
                 role = ''
 
-            if is_sensitive(self.action):
+            if self.is_sensitive():
                 clicked = self.action
             else:
                 clicked = None
@@ -727,8 +836,7 @@ class Button(renpy.display.layout.Window):
     def event(self, ev, x, y, st):
 
         def handle_click(action):
-            if self.style.activate_sound:
-                renpy.audio.music.play(self.style.activate_sound, channel="sound")
+            renpy.exports.play(self.style.activate_sound)
 
             rv = run(action)
 
@@ -745,11 +853,19 @@ class Button(renpy.display.layout.Window):
 
         # If we have a child, try passing the event to it. (For keyboard
         # events, this only happens if we're focused.)
-        if self.is_focused() or not (ev.type == pygame.KEYDOWN or ev.type == pygame.KEYUP):
+        if (not (ev.type in KEY_EVENTS)) or self.style.key_events:
             rv = super(Button, self).event(ev, x, y, st)
             if rv is not None:
                 return rv
 
+        if (self.keysym is not None) and (self.clicked is not None):
+            if map_event(ev, self.keysym):
+                return handle_click(self.clicked)
+
+        if (self.alternate_keysym is not None) and (self.alternate is not None):
+            if map_event(ev, self.alternate_keysym):
+                return handle_click(self.alternate)
+
         # If not focused, ignore all events.
         if not self.is_focused():
             return None
@@ -794,7 +910,6 @@ class Button(renpy.display.layout.Window):
 
         return None
 
-
     def set_style_prefix(self, prefix, root):
         if root:
             super(Button, self).set_style_prefix(prefix, root)
@@ -803,16 +918,24 @@ class Button(renpy.display.layout.Window):
         return ""
 
     def _tts_all(self):
-        return self._tts_common(alt(self.action))
+        rv = self._tts_common(alt(self.action))
+
+        if self.is_selected():
+            rv += " " + renpy.minstore.__("selected")
+
+        return rv
 
 # Reimplementation of the TextButton widget as a Button and a Text
 # widget.
+
+
 def TextButton(text, style='button', text_style='button_text',
                clicked=None, **properties):
 
-    text = renpy.text.text.Text(text, style=text_style) #@UndefinedVariable
+    text = renpy.text.text.Text(text, style=text_style)  # @UndefinedVariable
     return Button(text, style=style, clicked=clicked, **properties)
 
+
 class ImageButton(Button):
     """
     Used to implement the guts of an image button.
@@ -820,13 +943,13 @@ class ImageButton(Button):
 
     def __init__(self,
                  idle_image,
-                 hover_image = None,
-                 insensitive_image = None,
-                 activate_image = None,
-                 selected_idle_image = None,
-                 selected_hover_image = None,
-                 selected_insensitive_image = None,
-                 selected_activate_image = None,
+                 hover_image=None,
+                 insensitive_image=None,
+                 activate_image=None,
+                 selected_idle_image=None,
+                 selected_hover_image=None,
+                 selected_insensitive_image=None,
+                 selected_activate_image=None,
                  style='image_button',
                  clicked=None,
                  hovered=None,
@@ -842,15 +965,15 @@ class ImageButton(Button):
         selected_activate_image = selected_activate_image or activate_image
 
         self.state_children = dict(
-            idle_ = renpy.easy.displayable(idle_image),
-            hover_ = renpy.easy.displayable(hover_image),
-            insensitive_ = renpy.easy.displayable(insensitive_image),
-            activate_ = renpy.easy.displayable(activate_image),
+            idle_=renpy.easy.displayable(idle_image),
+            hover_=renpy.easy.displayable(hover_image),
+            insensitive_=renpy.easy.displayable(insensitive_image),
+            activate_=renpy.easy.displayable(activate_image),
 
-            selected_idle_ = renpy.easy.displayable(selected_idle_image),
-            selected_hover_ = renpy.easy.displayable(selected_hover_image),
-            selected_insensitive_ = renpy.easy.displayable(selected_insensitive_image),
-            selected_activate_ = renpy.easy.displayable(selected_activate_image),
+            selected_idle_=renpy.easy.displayable(selected_idle_image),
+            selected_hover_=renpy.easy.displayable(selected_hover_image),
+            selected_insensitive_=renpy.easy.displayable(selected_insensitive_image),
+            selected_activate_=renpy.easy.displayable(selected_activate_image),
             )
 
         super(ImageButton, self).__init__(None,
@@ -868,6 +991,7 @@ class ImageButton(Button):
 
 # This is used for an input that takes its focus from a button.
 class HoveredProxy(object):
+
     def __init__(self, a, b):
         self.a = a
         self.b = b
@@ -878,7 +1002,60 @@ class HoveredProxy(object):
             return self.b()
 
 
-class Input(renpy.text.text.Text): #@UndefinedVariable
+# The currently editable input value.
+current_input_value = None
+
+# Is the current input value active?
+input_value_active = False
+
+# The default input value to use if the currently editable value doesn't
+# exist.
+default_input_value = None
+
+# A list of input values that exist.
+input_values = [ ]
+
+# A list of inputs that exist in the current interaction.
+inputs = [ ]
+
+
+def input_pre_per_interact():
+    global input_values
+    global inputs
+    global default_input_value
+
+    input_values = [ ]
+    inputs = [ ]
+    default_input_value = None
+
+
+def input_post_per_interact():
+
+    global current_input_value
+    global input_value_active
+
+    for i in input_values:
+        if i is current_input_value:
+            break
+
+    else:
+
+        current_input_value = default_input_value
+
+        input_value_active = True
+
+    for i in inputs:
+
+        editable = (i.value is current_input_value) and input_value_active and i.value.editable
+
+        content = i.value.get_text()
+
+        if (i.editable != editable) or (content != i.content):
+            i.update_text(content, editable)
+            i.caret_pos = len(content)
+
+
+class Input(renpy.text.text.Text):  # @UndefinedVariable
     """
     This is a Displayable that takes text as input.
     """
@@ -891,6 +1068,7 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
     pixel_width = None
     default = u""
     edit_text = u""
+    value = None
 
     def __init__(self,
                  default="",
@@ -905,10 +1083,16 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
                  replaces=None,
                  editable=True,
                  pixel_width=None,
+                 value=None,
                  **properties):
 
         super(Input, self).__init__("", style=style, replaces=replaces, substitute=False, **properties)
 
+        if value:
+            self.value = value
+            changed = value.set_text
+            default = value.get_text()
+
         self.default = unicode(default)
         self.content = self.default
 
@@ -946,7 +1130,6 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
 
         self.update_text(self.content, self.editable)
 
-
     def _show(self):
         if self.default != self.content:
             self.content = self.default
@@ -1001,7 +1184,7 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
             if editable:
                 l = len(content)
                 self.set_text([self.prefix, content[0:self.caret_pos].replace("{", "{{"), edit_text, caret,
-                                            content[self.caret_pos:l].replace("{", "{{"), self.suffix])
+                               content[self.caret_pos:l].replace("{", "{{"), self.suffix])
             else:
                 self.set_text([self.prefix, content.replace("{", "{{"), self.suffix ])
 
@@ -1033,6 +1216,18 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
     def disable(self):
         self.update_text(self.content, False)
 
+    def per_interact(self):
+
+        global default_input_value
+
+        if self.value is not None:
+
+            inputs.append(self)
+            input_values.append(self.value)
+
+            if self.value.default and (default_input_value is None):
+                default_input_value = self.value
+
     def event(self, ev, x, y, st):
 
         self.old_caret_pos = self.caret_pos
@@ -1040,6 +1235,9 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
         if not self.editable:
             return None
 
+        if (ev.type == pygame.KEYDOWN) and (pygame.key.get_mods() & pygame.KMOD_LALT) and (not ev.unicode):
+            return None
+
         l = len(self.content)
 
         raw_text = None
@@ -1061,6 +1259,9 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
             if self.edit_text:
                 content = content[0:self.caret_pos] + self.edit_text + self.content[self.caret_pos:]
 
+            if self.value:
+                return self.value.enter()
+
             if not self.changed:
                 return content
 
@@ -1110,6 +1311,7 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
             raw_text = ev.text
 
         elif ev.type == pygame.KEYDOWN:
+
             if ev.unicode and ord(ev.unicode[0]) >= 32:
                 raw_text = ev.unicode
             elif renpy.display.interface.text_event_in_queue():
@@ -1149,12 +1351,15 @@ class Input(renpy.text.text.Text): #@UndefinedVariable
 
         return rv
 
+
 # A map from adjustment to lists of displayables that want to be redrawn
 # if the adjustment changes.
 adj_registered = { }
 
 # This class contains information about an adjustment that can change the
 # position of content.
+
+
 class Adjustment(renpy.object.Object):
     """
     :doc: ui
@@ -1168,7 +1373,7 @@ class Adjustment(renpy.object.Object):
 
     """
 
-    def __init__(self, range=1, value=0, step=None, page=0, changed=None, adjustable=None, ranged=None): #@ReservedAssignment
+    def __init__(self, range=1, value=0, step=None, page=None, changed=None, adjustable=None, ranged=None):  # @ReservedAssignment
         """
         The following parameters correspond to fields or properties on
         the adjustment object:
@@ -1217,7 +1422,6 @@ class Adjustment(renpy.object.Object):
             any bars and viewports that use the adjustment.
          """
 
-
         super(Adjustment, self).__init__()
 
         if adjustable is None:
@@ -1251,7 +1455,7 @@ class Adjustment(renpy.object.Object):
         if self.ranged:
             self.ranged(self)
 
-    range = property(get_range, set_range) #@ReservedAssignment
+    range = property(get_range, set_range)  # @ReservedAssignment
 
     def get_page(self):
         if self._page is not None:
@@ -1301,6 +1505,16 @@ class Adjustment(renpy.object.Object):
 
         return None
 
+    def update(self):
+        """
+        Updates things that depend on this adjustment without firing the
+        changed handler.
+        """
+
+        for d in adj_registered.setdefault(self, [ ]):
+            renpy.display.render.redraw(d, 0)
+
+
 class Bar(renpy.display.core.Displayable):
     """
     Implements a bar that can display an integer value, and respond
@@ -1312,17 +1526,17 @@ class Bar(renpy.display.core.Displayable):
     def after_upgrade(self, version):
 
         if version < 1:
-            self.adjustment = Adjustment(self.range, self.value, changed=self.changed) # E1101
+            self.adjustment = Adjustment(self.range, self.value, changed=self.changed)  # E1101
             self.adjustment.register(self)
-            del self.range # E1101
-            del self.value # E1101
-            del self.changed # E1101
+            del self.range  # E1101
+            del self.value  # E1101
+            del self.changed  # E1101
 
         if version < 2:
             self.value = None
 
     def __init__(self,
-                 range=None, #@ReservedAssignment
+                 range=None,  # @ReservedAssignment
                  value=None,
                  width=None,
                  height=None,
@@ -1410,10 +1624,17 @@ class Bar(renpy.display.core.Displayable):
             if redraw is not None:
                 renpy.display.render.redraw(self, redraw)
 
+        xminimum = self.style.xminimum
+        yminimum = self.style.yminimum
+
+        if xminimum is not None:
+            width = max(width, xminimum)
+            height = max(height, yminimum)
+
         # Store the width and height for the event function to use.
         self.width = width
         self.height = height
-        range = self.adjustment.range #@ReservedAssignment
+        range = self.adjustment.range  # @ReservedAssignment
         value = self.adjustment.value
         page = self.adjustment.page
 
@@ -1518,7 +1739,6 @@ class Bar(renpy.display.core.Displayable):
 
         return rv
 
-
     def focus(self, default=False):
         super(Bar, self).focus(default)
         self.set_transform_event("hover")
@@ -1526,7 +1746,6 @@ class Bar(renpy.display.core.Displayable):
         if not default:
             run(self.hovered)
 
-
     def unfocus(self, default=False):
         super(Bar, self).unfocus()
         self.set_transform_event("idle")
@@ -1546,7 +1765,7 @@ class Bar(renpy.display.core.Displayable):
         if self.hidden:
             return None
 
-        range = self.adjustment.range #@ReservedAssignment
+        range = self.adjustment.range  # @ReservedAssignment
         old_value = self.adjustment.value
         value = old_value
 
@@ -1558,12 +1777,15 @@ class Bar(renpy.display.core.Displayable):
         grabbed = (renpy.display.focus.get_grab() is self)
         just_grabbed = False
 
+        ignore_event = False
+
         if not grabbed and map_event(ev, "bar_activate"):
-            renpy.display.tts.speak("activate")
+            renpy.display.tts.speak(renpy.minstore.__("activate"))
             renpy.display.focus.set_grab(self)
             self.set_style_prefix("selected_hover_", True)
             just_grabbed = True
             grabbed = True
+            ignore_event = True
 
         if grabbed:
 
@@ -1575,12 +1797,14 @@ class Bar(renpy.display.core.Displayable):
                 decrease = "bar_left"
 
             if map_event(ev, decrease):
-                renpy.display.tts.speak("decrease")
+                renpy.display.tts.speak(renpy.minstore.__("decrease"))
                 value -= self.adjustment.step
+                ignore_event = True
 
             if map_event(ev, increase):
-                renpy.display.tts.speak("increase")
+                renpy.display.tts.speak(renpy.minstore.__("increase"))
                 value += self.adjustment.step
+                ignore_event = True
 
             if ev.type in (pygame.MOUSEMOTION, pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN):
 
@@ -1603,27 +1827,37 @@ class Bar(renpy.display.core.Displayable):
                     else:
                         value = 0
 
+                ignore_event = True
+
             if isinstance(range, int):
                 value = int(value)
 
             if value < 0:
+                renpy.display.tts.speak("")
                 value = 0
 
             if value > range:
+                renpy.display.tts.speak("")
                 value = range
 
         if invert:
             value = range - value
 
         if grabbed and not just_grabbed and map_event(ev, "bar_deactivate"):
-            renpy.display.tts.speak("deactivate")
+            renpy.display.tts.speak(renpy.minstore.__("deactivate"))
             self.set_style_prefix("hover_", True)
             renpy.display.focus.set_grab(None)
+            ignore_event = True
 
         if value != old_value:
-            return self.adjustment.change(value)
+            rv = self.adjustment.change(value)
+            if rv is not None:
+                return rv
 
-        return None
+        if ignore_event:
+            raise renpy.display.core.IgnoreEvent()
+        else:
+            return None
 
     def set_style_prefix(self, prefix, root):
         if root:
@@ -1637,9 +1871,9 @@ class Bar(renpy.display.core.Displayable):
         if self.value is not None:
             alt = self.value.alt
         else:
-            alt = "Bar"
+            alt = ""
 
-        return self._tts_common(alt)
+        return self._tts_common(alt) + renpy.minstore.__("bar")
 
 
 class Conditional(renpy.display.layout.Container):
@@ -1688,6 +1922,7 @@ class TimerState(renpy.python.RevertableObject):
     started = False
     next_event = None
 
+
 class Timer(renpy.display.layout.Null):
 
     __version__ = 1
@@ -1731,7 +1966,6 @@ class Timer(renpy.display.layout.Null):
         else:
             self.state = TimerState()
 
-
     def event(self, ev, x, y, st):
 
         state = self.state
@@ -1813,13 +2047,13 @@ class MouseArea(renpy.display.core.Displayable):
             run_unhovered(self.hovered)
             run(self.unhovered)
 
+
 class OnEvent(renpy.display.core.Displayable):
     """
     This is a displayable that runs an action in response to a transform
     event. It's used to implement the screen language on statement.
     """
 
-
     def __init__(self, event, action=[ ]):
         """
         `event`
@@ -1834,6 +2068,12 @@ class OnEvent(renpy.display.core.Displayable):
         self.event_name = event
         self.action = action
 
+    def _handles_event(self, event):
+        if self.event_name == event:
+            return True
+        else:
+            return False
+
     def set_transform_event(self, event):
         if event == self.event_name:
             run(self.action)
